# NoccyLabs Interactive Shell

This is a self-contained asynchronous interactive shell for PHP apps.

## Key features

* Zero dependencies, all code isolated in namespace. Should play well with
  just about everything.
* Readline-like functionality, without readline. Cursor control, history
  scrollback and more.
* Context-centered, using a stack on which you can push contexts to dive
  deeper, or pop contexts to return to the top. All active contexts are
  refreshed constantly, making it possible to have the session context
  report errors while a query session is at the top of the stack.
* Commands decopled from contexts by default (it is still fully possible to
  create self-contained contexts) making it easy to share logic between 
  multiple contxts.
* Fully asynchronous, your contexts are updated frequently to poll sockets,
  report status and other useful things. Meanwhile, your prompt isn't going 
  anywhere!
* The output is decoupled, opening up for displaying the shell on a connected
  LCD display, dot-matrix printer+keyboard, or maybe websockets?
* ANSI color support (16 colors) including styles (bold, dim, underline,
  italic) through tags.
* Partially detached input/output system, making it possible to create even
  more interactive experiences in the future without major updates to existing
  user code.
* Typed settings storage with obsevers, through the OptionSet class. Get
  notified as an option is updated by observing the setting.



## Minimal example

```php
<?php

require_once __DIR__."/vendor/autoload.php";

use NoccyLabs\InteractiveShell\Command\Commands\HelpCommand;
use NoccyLabs\InteractiveShell\Command\Commands\QuitCommand;
use NoccyLabs\InteractiveShell\Shell;
use NoccyLabs\InteractiveShell\Context\ShellContext;

$shell = new Shell();

$context = new ShellContext();
$context->addCommand(new HelpCommand());
$context->addCommand(new QuitCommand());

$shell->run($context);
```

## Custom contexts

```php

class ConnectionContext extends ShellContext
{
    protected function build(): void
    {
        // add your commands and stuff here
    }

    public function update(): void
    {
        // do work here, but don't take too long!
    }

    public function contextAttached(): void
    {
        // update the prompt or so
    }

    public function contextDetached(): void
    {
        // stop stuff if needed
    }
}
```

## Custom commands

```php
class OtherCommand extends ShellCommand
{
    protected function build(): void
    {
        $this->setCommandName("other");
        $this->setDescription("Enter other context");
    }

    public function invoke(array $args, ShellContext $context)
    {
        $context->getShell()->pushContext(new OtherContext());
    }
}
```


## Known bugs

* As everything goes through ConsoleOutput::write(), tags in commands are
  escaped. So it would look silly if someone decided to put `<bold>` or
  a color tag in the input. So this is more of a cosmetic edge case of
  unsupproted input.
