# Interactive Shell Tutorial

This is a library that help you build powerful and colorful interactive shells into just about any PHP application. Or at least CLI apps for now, but the library is built such that the input and output could easily take place over WebSockets.

> **NOTE:** The code examples in this tutorial are written for brevity rather than to comply with coding styles. If you copy and paste the examples, at least have the decency to run it through a code beautifier! 

> **NOTE2:** When you encounter "`// ...`" in the examples, this means leave whatever is already there and append/prepend the rest. As mentioned in the previous block, the examples are written for brevity.

## Concepts

The shell will keep on running for as long as there is at least one context on its context stack. You will soon understand what this means.

### But what is a shell?

A shell is essentially a REPL, or a read-evaluate-print loop that accepts command strings, invokes the requested command, and returns the result. Using the library, these commands can be executed in at least 3 different ways that have an impact on how the output is presented:

1. Commands can execute as in, in the commands `invoke()` method. The output is buffered and will be presented *after* the method returns, and as such can show no progress indications. The output is parsed however, so you can use tags to provide colors and markup.
2. Commands can execute asynchronously, by having the `invoke()` method update the context such that calls to the contexts `update()` method advances the operation until completed. The output is buffered, but dumped as chunks and as such the output from each of the update calls will be presented *after* the update method returns. The output is again parsed.
3. Commands can execute in the foreground by requesting exclusivity, by passing a callback to the `Shell::exclusive()` method. This callback will execute unbuffered, and output is presented *immediately* as it is written. This output is not parsed, so markup is not supported.

### What is a context?

A context is what is currently being operated upon. It holds the commands and any additional logic needed to implement the specific functionality. Imagine an FTP client (If you were born after 2000, ask your parents or Google). When you initially open it up you are able to change settings and connect to a server. Once you are connected you can probably access the settings too, but you can also put files, get files, list files and all that. In this case you would have one context that would handle making connections, and one that handles the actual connected session. 

### What is a command?

Come on, really? .. Well, a command is something that can be invoked by using its command name. The library wraps all commands in classes, meaning you commands can have state. There is however no built-in logic to perform updates on commands, meaning they are only active when invoked. Any logic that need to work in the background should go in the context.

## Getting started

It all starts with the shell:

```php
<?php

use NoccyLabs\InteractiveShell\Shell;
use NoccyLabs\InteractiveShell\Context\ShellContext;

class MyShellRunner {

    private Shell $shell;

    private ShellContext $context;

    public function startShell(): void {
        $this->context = new ShellContext();
        $this->shell = new Shell();
        $this->shell->run($this->context);
    }

}
```

Go ahead and call on `startShell()`. The first thing you will notice is a colorful prompt, and that your input is echoed. The second thing you will notice is that you can't quit, and that `q!` isn't working! ;) Don't worry, there is a debugging built-in for that. Just type `?quit`, and we will get around to fixing the problem!

```php

// ...
use NoccyLabs\InteractiveShell\Command\Commands\QuitCommand;

class MyShellRunner {

    // ...

    public function startShell(): void {
        $this->context = new ShellContext();
        $this->context->addCommand(new QuitCommand());
        // ...
    }

}
```

Sweet! You can now use `quit` to exit! If you go ahead and add the `HelpCommand` you will also be able to pull up a colorful help using `help`. But we really need to put it all in a custom context:

```php
<?php

use NoccyLabs\InteractiveShell\Shell;
use NoccyLabs\InteractiveShell\Context\ShellContext;
use NoccyLabs\InteractiveShell\Command\Commands\QuitCommand;
use NoccyLabs\InteractiveShell\Command\Commands\HelpCommand;

class RootContext extends ShellContext {

    protected function build(): void {
        $this->addCommand(new HelpCommand());
        $this->addCommand(new QuitCommand());
    }

}

class MyShellRunner {

    private Shell $shell;

    private ShellContext $context;

    public function startShell(): void {
        $this->context = new RootContext();
        $this->shell = new Shell();
        $this->shell->run($this->context);
    }

}
```

All our logic is now located in `RootContext` and we can start making use of some of the provided magic. First is the `update()` method; this method is invoked once every loop so you should not do any heavy lifting in here. Next are the `contextAttached()` and `contextDetached()` methods that are invoked whenever the context is becoming the primary context, and when it is no longer primary. Let's add update and contextAttached to the context:

```php
<?php

// ...

class RootContext extends ShellContext {

    // ...

    public function update(): void {
        // select some streams, read some sockets, do some wor
    }

    public function contextAttached(): void {
        $this->getShell()->setPrompt("<fg=cyan>O<fg=magenta>M<fg=blue>G</>> ");
    }

}
```

Bazinga! Your prompt now looks like unicorn puke! But at least you learned 1) how to update the prompt when your context is activated, and 2) that you can you tags in the prompt! 

Let's investigate these tags a bit closer by adding a custom command:

```php

// ...
use NoccyLabs\InteractiveShell\Command\ShellCommand;

class TestCommand extends ShellCommand {

    protected function build(): void {
        $this->setName("test");
        $this->setDescription("This is a test command");
    }

    public function invoke(array $args, ShellContext $context): void {
        // note how we can use echo and friends here, since output is buffered
        echo "Hello World!\n";
        echo "<options=bold>This text is bold</options=bold> but <bold>so is this!</>\n";
        echo "There is also <dim>dim</>, <italic>italic</>, <underline>underline</>!\n";
        echo "<fg=red>Red text</> <fg=black;bg=yellow>Black on Yellow</>\n";
    }

}

class RootContext extends ShellContext {

    protected function build(): void {
        // ...
        $this->addCommand(new TestCommand());
    }

}
```

Go ahead! Try the `test` command out! As you see we get the context as well as any arguments passed, giving access to the essentials. As for the tags, the anal XML author inside you may be screaming about closing tags, but tell them to put a sock in it. The formatting tags don't mark up a text block, so there is no need for closing tags. Instead they insert ANSI control sequences that does the formatting.

Here, this may not be a cheat-sheet printed on large size coated paper, but it should do the trick:

```
   __ style/color off, if no attributes reset color and style
  |
<[/][ATTR[,ATTR..]]>
     |_____|
     |
   [KEY=]VALUE
     |     |__ color or comma-separated list of styles
     |
     |__ fg: Set foreground color
     |__ bg: Set backgound color
     |__ options: Set style
```

Another useful debugging built-in is the `?context` command. It will display the context stack, which right now is completely useless since we only have a single context. Let's make a new `TestContext` and move the command we just created over to it. We can then add a `test` command to the `RootContext` that pushes the new context. While we're at it, we might as well add a `back` command!

```php
// ...
use NoccyLabs\InteractiveShell\Command\Commands\Command;

class TestContext extends ShellContext {

    protected function build(): void {
        $this->addCommand(new HelpCommand());
        $this->addCommand(new QuitCommand());
        $this->addCommand(new Command("back", "Exit the test mode", 
            function (array $args, ShellContext $context) {
                $context->getShell()->popContext();
            })
        );
    }

    public function contextAttached(): void {
        $this->getShell()->setPrompt("test> ");
    }

}

class RootContext extends ShellContext {

    protected function build(): void {
        // ..
        $this->addCommand(new Command("test", "Enter test mode", 
            function (array $args, ShellContext $context) {
                $context->getShell()->pushContext(new TestContext());
            })
        );
    }

}
```

As you can see, we added the `help` and `quit` commands to our new context, as well as the `back` command. There are no global commands, so anything that should be available must be explicitly added. You can also see that functions can be closures, although they are as previously mentioned wrapped up in a class. The `Command` class provides a handy "hollow" class to hold closures or defer the call to a context or other object.


### Complete example

```php
<?php

use NoccyLabs\InteractiveShell\Shell;
use NoccyLabs\InteractiveShell\Context\ShellContext;
use NoccyLabs\InteractiveShell\Command\Commands\QuitCommand;
use NoccyLabs\InteractiveShell\Command\Commands\HelpCommand;
use NoccyLabs\InteractiveShell\Command\Commands\Command;
use NoccyLabs\InteractiveShell\Command\ShellCommand;

class TestCommand extends ShellCommand {

    protected function build(): void {
        $this->setCommandName("test");
        $this->setDescription("This is a test command");
    }

    public function invoke(array $args, ShellContext $context): void {
        // note how we can use echo and friends here, since output is buffered
        echo "Hello World!\n";
        echo "<options=bold>This text is bold</options=bold> but <bold>so is this!</>\n";
        echo "There is also <dim>dim</>, <italic>italic</>, <underline>underline</>!\n";
        echo "<fg=red>Red text</> <fg=black;bg=yellow>Black on Yellow</>\n";
    }

}

class TestContext extends ShellContext {

    protected function build(): void {
        $this->addCommand(new HelpCommand());
        $this->addCommand(new QuitCommand());
        $this->addCommand(new TestCommand());
        $this->addCommand(new Command("back", "Exit the test mode", 
            function (array $args, ShellContext $context) {
                $context->getShell()->popContext();
            })
        );
    }

    public function contextAttached(): void {
        $this->getShell()->setPrompt("test> ");
    }

}

class RootContext extends ShellContext {

    protected function build(): void {
        $this->addCommand(new HelpCommand());
        $this->addCommand(new QuitCommand());
        $this->addCommand(new Command("test", "Enter test mode", 
            function (array $args, ShellContext $context) {
                $context->getShell()->pushContext(new TestContext());
            })
        );
    }

    public function contextAttached(): void {
        $this->getShell()->setPrompt("<fg=cyan>O<fg=magenta>M<fg=blue>G</>> ");
    }

}

class MyShellRunner {

    private Shell $shell;

    private ShellContext $context;

    public function startShell(): void {
        $this->context = new RootContext();
        $this->shell = new Shell();
        $this->shell->run($this->context);
    }

}

$shell = new MyShellRunner();
$shell->startShell();

```
