<?php

namespace NoccyLabs\InteractiveShell\Context;

use NoccyLabs\InteractiveShell\Command\ShellCommand;
use NoccyLabs\InteractiveShell\Shell;

class ShellContext
{

    private array $commands = [];

    /** @var array<ShellContext> */
    private array $extensions = [];

    private Shell $shell;

    public function __construct()
    {
        $this->build();
    }

    public function setShell(Shell $shell)
    {
        $this->shell = $shell;
    }

    public function getShell(): Shell
    {
        return $this->shell;
    }

    protected function build(): void
    {

    }

    public function addContextExtension(ShellContext $extension)
    {
        if (in_array($extension, $this->extensions)) return;
        $extension->contextAttached();
        $this->extensions[] = $extension;
    }

    public function removeContextExtension(ShellContext $extension)
    {
        if (!in_array($extension, $this->extensions)) return;
        $extension->contextDetached();
        $this->extensions = array_values(
            array_filter(
                $this->extensions,
                fn($v) => $v != $extension
            )
        );
    }

    public function getContextExtensions(): array
    {
        return $this->extensions;
    }

    public function addCommand(ShellCommand $command, string $commandName=null)
    {
        $c = $commandName ?? $command->getCommandName();

        $this->commands[$c] = $command;
    }

    public function findCommand(string $command): ?ShellCommand
    {
        $commandList = $this->commands;
        foreach ($this->extensions as $extension) {
            $commandList = array_merge(
                $commandList,
                $extension->commands
            );
        }
        if (array_key_exists($command, $commandList)) {
            return $commandList[$command];
        }

        $match = [];
        foreach ($commandList as $key=>$handler) {
            if (str_starts_with($key, $command)) {
                $match[$key] = $handler;
            }
        }

        if (count($match) == 1) {
            return reset($match);
        } elseif (count($match) > 1) {
            printf("Ambiguous command %s (matches %s)\n", $command, join(",", array_keys($match)));
        }

        return null;
    }

    public function findAllCommands(): array
    {
        $commands = $this->commands;
        foreach ($this->extensions as $extension) {
            $commands = array_merge($commands, $extension->commands);
        }
        return $commands;
    }

    public function executeCommand(string $command, array $arguments)
    {
        $c = $this->commands[$command] ?? null;
        if (!$c) {
            echo "No such command: {$command}\n";
            return;
        }
        $c->invoke($arguments, $this);
        $this->afterCommand();
    }

    /**
     * Called when the context becomes active, i.e. the root context is activated
     * or a child context is activated from another context.
     * 
     */
    public function contextAttached(): void
    {

    }

    /**
     * Called when the context is detached, i.e. another context is about to be
     * activated. Stop any foreground voodoo here.
     */
    public function contextDetached(): void
    {

    }

    public function afterCommand(): void
    {
    }

    public function update(): void
    {

    }

}
