<?php

namespace NoccyLabs\InteractiveShell;

use NoccyLabs\InteractiveShell\Context\ShellContext;
use NoccyLabs\InteractiveShell\Event\CommandExecuteEvent;
use NoccyLabs\InteractiveShell\Event\Event;
use NoccyLabs\InteractiveShell\Event\ShellContextUpdatedEvent;
use NoccyLabs\InteractiveShell\Input\ConsoleInput;
use NoccyLabs\InteractiveShell\Input\InputBuffer;
use NoccyLabs\InteractiveShell\Input\InputInterface;
use NoccyLabs\InteractiveShell\Output\ConsoleFormatter;
use NoccyLabs\InteractiveShell\Output\ConsoleOutput;
use NoccyLabs\InteractiveShell\Output\OutputInterface;

class Shell
{

    private array $contextStack = [];

    private array $eventListeners = [];

    private array $eventQueue = [];

    private ?InputInterface $input = null;

    private ?OutputInterface $output = null;

    private InputBuffer $buffer;

    private ConsoleFormatter $formatter;

    private bool $shellRunning = false;

    public function __construct()
    {
        $this->buffer = new InputBuffer();
    }

    public function pushContext(ShellContext $context)
    {
        // Make the context aware of the shell instance.
        $context->setShell($this);

        $c = end($this->contextStack);
        if ($c) {
            $c->contextDetached();
        } else {
            $c = null;
        }
        $context->contextAttached();

        $this->contextStack[] = $context;

        $this->sendEvent(new ShellContextUpdatedEvent($context, $c));
        // TODO notify event that context has changed

    }

    public function popContext(): ?ShellContext
    {
        $p = array_pop($this->contextStack);
        if ($p) {
            $p->contextDetached();
        } else {
            $p = null;
        }

        $c = end($this->contextStack);
        if ($c) {
            $c->contextAttached();
            $this->sendEvent(new ShellContextUpdatedEvent($c, $p));
        } else {
            $this->exitShell();
        }

        // TODO notify event that context has changed

        return $p;
    }

    public function addListener(string $event, callable $handler, bool $prepend=false)
    {
        if (!array_key_exists($event, $this->eventListeners)) {
            $this->eventListeners[$event] = [];
        }
        
        if (in_array($handler, $this->eventListeners[$event], true)) {
            return;
        }

        if ($prepend) {
            array_unshift($this->eventListeners[$event], $handler);
        } else {
            array_push($this->eventListeners[$event], $handler);
        }
    }

    public function removeListener(string $event, callable $handler)
    {
        if (!array_key_exists($event, $this->eventListeners)) {
            $this->eventListeners[$event] = [];
        }

        $this->eventListeners[$event] = array_values(
            array_filter(
                $this->eventListeners[$event],
                fn($e) => $handler !== $e
            )
        );
    }

    public function postEvent(Event $event)
    {
        $this->eventQueue[] = $event;
    }

    public function sendEvent(Event $event)
    {
        $fqcn = get_class($event);
        if (!array_key_exists($fqcn, $this->eventListeners)) {
            return;
        }
        foreach ($this->eventListeners as $listener) {
            if (call_user_func($listener, $event) === true) {
                break;
            }
        }
    }

    public function dispatchEvents()
    {
        $l = count($this->eventQueue);
        while ($l-- > 0) {
            $e = array_shift($this->eventQueue);
            if ($e) {
                $this->sendEvent($e);
            }
        }
    }

    public function executeBuffer(string $buffer)
    {
        $this->output->flushPrompt();

        if (trim($buffer)) {
            $this->buffer->addHistory($buffer);
        }
        $commandLine = array_values(array_filter(str_getcsv($buffer, " ")));

        $this->execute($commandLine);
    }

    public function execute(array $commandLine)
    {
        $args = $commandLine;
        $command = array_shift($args);
        if (!$command) {
            return;
        }

        $handler = $this->findBuiltInCommand($command);
        if ($handler) {
            call_user_func($handler, $args);
            // TODO Turn this into an event
            $context->afterCommand();
            return;
        }

        $context = end($this->contextStack);
        $handler = $context->findCommand($command);
        if ($handler) {
            $handler->invoke($args, $context);
            // TODO Turn this into an event
            $context->afterCommand();
            return;
        }

        echo "Bad command\n";

    }

    public function findBuiltInCommand(string $command)
    {
        // Check for some built-ins
        switch ($command) {
            case "?quit": return function () {
                    $this->shellRunning = false;
                };
            case '?context': return function () {
                    foreach ($this->contextStack as $n=>$context) {
                        echo ($n==0?" \u{2500}":"  ").
                            ($n>1?str_repeat("  ",$n-1):"").
                            ($n>0?"\u{2514}\u{2500}":"").
                            "<fg=yellow".($context==end($this->contextStack)?";bold":"").">".get_class($context)."</>\n";
                    }
                };
            case "?version": return function () {
                    echo "InteractiveShell 0.1.0\n";
                };
        }
    }

    public function run(ShellContext $context)
    {
        
        $this->formatter = new ConsoleFormatter();

        $this->input = new ConsoleInput($this, $this->buffer);
        $this->output = new ConsoleOutput($this->formatter, $this->buffer);
    
        ob_start();

        $this->pushContext($context);
        
        $this->shellRunning = true;

        while ($this->shellRunning) {

            // Update the input
            $this->input->update();

            // Dispatch any posted events
            $this->dispatchEvents();

            // Update the context if we have one
            foreach ($this->contextStack as $context) {
                $context->update();
            }

            $content = ob_get_contents();

            ob_end_clean();

            if ($content) {
                $this->output->writeFormatted($content);
            }

            // If the input buffer has been updated, we need to update the prompt
            if ($this->buffer->updated() || $content) {
                $this->output->redrawPrompt();
            }

            usleep(10000);

            ob_start();
        }

        ob_end_clean();
        
        $this->output->erasePrompt();

    }

    public function exclusive(callable $stub, array $args=[])
    {
        if ($content = ob_get_contents()) {
            $this->output->writeFormatted($content);
        }
        ob_end_clean();

        echo "\n";
        call_user_func($stub, ...$args);

        ob_start();
    }

    public function setPrompt(string $prompt): void
    {
        $this->buffer->setPrompt($prompt);
    }

    public function exitShell()
    {
        $this->shellRunning = false;
    }
}
